import sounddevice
import itertools
import time

from pydrummer.settings import PlaybackSettings

SECONDS_PER_MIN = 60.0
MIN_STEP_TIME = 0.01    # TODO: Figure out a way to not have to cut off notes at the
                        # beginning of the next step. This is will make it so we don't
                        # hit this minimum so often, which will give us a less choppy
                        # sound.


class Player(object):
    def __init__(self):
        self.settings = PlaybackSettings(bpm=120, time_signature=(4,4))

    def play(self, song=None, clip=None, sequence=None, sound=None, loops=-1):
        if sound:
            self.play_sound(sound=sound, blocking=True)
            print('')
        if sequence:
            self.play_sequence(sequence=sequence, loops=loops)
        if clip:
            self.play_clip(clip=clip, loops=loops)
        if song:
            self.play_song(song=song, loops=loops)

    def play_sound(self, sound, blocking=False):
        print('{}'.format(sound.name), end=' ', flush=True)
        sounddevice.play(sound.data_array, sound.sample_rate, blocking=blocking)

    # TODO: Improve accuracy of calculatation for when to play the next beat
    # Maybe try realigning every bar - use absolute time?
    def play_loop(self, sounds, step_time_s, loops=-1):
        """Play a note for every step in the clip

        Repeat for the specified number of loops. If loop is -1 then play continuously.
        """
        steps = len(sounds)
        loop_count = 0
        for i in itertools.cycle(range(steps)):
            now = time.time()
            # play the sound
            self.play_sound(sounds[i])
            # sleep for the remainding amount of time
            elapsed = time.time() - now
            time.sleep(max(MIN_STEP_TIME, step_time_s - elapsed))
            if i == steps - 1:
                loop_count += 1
                print('')
            if loop_count == loops:
                return

    def play_sequence(self, sequence, step_time_s=0.5, loops=-1):
        bar_time_s = (SECONDS_PER_MIN / self.settings.bpm) * self.settings.beats_per_measure
        step_time_s = bar_time_s / sequence.steps
        self.play_loop(sounds=sequence.get_sounds(), step_time_s=step_time_s, loops=loops)

    def play_clip(self, clip, loops=-1):
        self.play_sequence(sequence=clip.mix, loops=loops)

    def play_song(self, song, loops=-1):
        for clip in song.clips:
            self.play_clip(clip=clip, loops=loops)



from pydrummer.models.mix import Mix
from pydrummer.models.sequence import Sequence


class Clip(object):
    """A Clip maintains Sequences and the output Mix sequence.

    A Clip manages the Sequences used during the construction of a rhythm and
    maintains the Mix, the output rhythm created by merging all the Sequences
    together.

    A Sequence is created for every instrument in the instruments dictionary.
    Both the Mix and the Sequences have the same number of steps.
    """

    def __init__(self, instruments, steps):
        self.steps = steps
        self.sequences = {}
        for name, fpath in instruments.items():
            self.sequences[name] = Sequence(name=name, fpath=fpath, steps=self.steps)
        self.mix = Mix(steps=self.steps)

    def add_pattern(self, name, pattern):
        """Add a pattern to the Sequence with the given name.

        Each time a Sequence is updated, we merge it with our Mix sequence so
        that it's ready to be played back at any time.
        """
        self.sequences[name].set_pattern(pattern)
        print('{pattern} ({name})'.format(name=name, pattern=self.sequences[name].pattern))
        self.mix.merge(self.sequences[name])



from pydrummer.models.sound import Sound


class Mix(object):
    """A Mix is a single multi-sound pattern created by combing Sequences.

    A Mix is used for rhythm playback.

    TODO: Look into merging steps into one big wave file where notes can
    continue to play until they finish instead of cutting every sound off at
    each step
    """

    def __init__(self, steps):
        self.steps = steps
        self.sounds = [Sound()] * steps
        self.pattern = [0] * steps

    def merge(self, sequence):
        """ Merge notes from the given sequence into the mix

        If the pattern in the given sequence has a note with the value of zero
        we do nothing and move on to the next note. If the note is one we merge
        it into the mix. We either merge the note into an existing note by
        mixing the two sounds together or we copy it into an empty slot in our
        pattern.
        """
        for i in range(self.steps):
            if sequence.pattern[i] == 0:  # nothing to copy or mix
                continue
            if self.pattern[i] == 0:  # copy note
                self.pattern[i] = 1
                self.sounds[i] = Sound(sequence.sound.name, sequence.sound.file_path)
                self.sounds[i].name = sequence.sound.name
            else:  # mix notes together
                self.sounds[i].mix(sequence.sound)
                self.sounds[i].name = 'mix'

    def get_sounds(self):
        return self.sounds



import numpy as np
import soundfile


class Sound(object):
    """A Sound is pre-recorded data that can be heard when it reaches our ears.

    A Sound can also be mixed with other Sounds to create a new Sound. This is
    how we create the effect of hearing two notes playing back at the same time.

    We read in the sound file to get its data and sample rate to be used during
    playback. When another sound is mixed in we continue to use the original
    sample_rate.
    """

    def __init__(self, name='-', fpath=''):
        self.name = name
        self.file_path = fpath
        if not fpath:
            self.data = []
            self.sample_rate = 44100
        else:
            self.data, self.sample_rate = soundfile.read(fpath, frames=100000, fill_value=0)
        self.data_array = np.array(self.data)

    def mix(self, sound):
        # TODO: Figure out how to mix soundfiles of different shapes
        # TODO: Fix sounddevice.PortAudioError: Error opening OutputStream:
        # Invalid number of channels [PaErrorCode -9998] - Hard to repro, try
        # playing clips with many different sequences with overlapping sounds
        # and try different wave files
        # TODO: Actually research how to mix sound data together to make this
        # sound better
        self.data_array = np.column_stack([self.data, sound.data])



from pydrummer.models.clip import Clip


class Song():
    """A Song is a container for Clips.
    """

    def __init__(self):
        self.clips = []

    def add_clip(self, instruments, steps=16):
        clip = Clip(instruments=instruments, steps=steps)
        self.clips.append(clip)



from pkg_resources import resource_string, resource_listdir, resource_filename
import os.path

"""Add your drumkits here.

DRUMKITS maps the name of a drumkit to its samples directory.
"""
DRUMKITS = {
    'tr808': 'samples/808',
}

"""Sample wave files for each drumkit provided in DRUMKITS.

e.g. To access the 'kick' sample in the 'tr808' drumkit:
kick_filepath = SAMPLES['tr808']['kick']
"""
SAMPLES = {
    name: {
        os.path.splitext(sample)[0]: resource_filename(__name__, os.path.join(path, sample))
        for sample in resource_listdir(__name__, path) if sample.endswith('.wav')
    }
    for name, path in DRUMKITS.items()
}


from setuptools import setup, find_packages

setup(
    name='pydrummer',
    version='0.1.dev0',
    description='a minimal drum machine',
    url='http://github.com/allieoop/pydrummer',
    author='Allie Crevier',
    author_email='okay@allie.is',
    license='MIT',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'numpy',
        'sounddevice',
        'soundFile'
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
    ],
)



from pydrummer.models.sound import Sound


class Sequence(object):
    """A Sequence is a series of notes defined by a pattern and a Sound.

    A pattern is an array of zeros and ones and has a length equal to the
    number of steps in our playback window.
    """

    def __init__(self, name, fpath, steps):
        self.name = name
        self.sound = Sound(name, fpath)
        self.steps = steps
        self.pattern = [0] * steps

    def set_pattern(self, pattern):
        """Set the pattern.

        If the length of given pattern is smaller than the length of the pattern
        in our Sequence, loop throught it again an continue setting elements
        until you reach the end.
        """
        for i in range(len(self.pattern)):
            self.pattern[i] = pattern[i % len(pattern)]

    def get_sounds(self):
        sounds = []
        for i in range(len(self.pattern)):
            if self.pattern[i] == 1:
                sounds.append(self.sound)
            else:
                sounds.append(Sound())
        return sounds
